

def time_convert(arr):
    num = int(arr[1])
    units = arr[2]
    if units == 'day':
        units = 'days'

    gran_order = [['start', 0], ['seconds', 59], ['minutes', 3599], ['hours', 86399],
                  ['days', 2591999], ['months', 31103999], ['years', 31104000000]]
    end_div = [['start', 0], ['seconds', 1], ['minutes', 60], ['hours', 3600], ['days', 86400],
               ['months', 2592000], ['years', 31104000]]
    all_units = [k[0] for k in gran_order]
    starting_point = all_units.index(units)
    starting_value = (gran_order[starting_point - 1][1] + 1) * num
    i = 0
    for i in range(starting_point, len(gran_order)):
        if gran_order[i][1] < starting_value:
            continue
        else:
            break
    conv = starting_value // end_div[i][1]
    un = end_div[i][0]

    if conv == 1:
        un = un[:-1]
    return str(conv), ' ' + un


def stretch_image_size(image_width, image_height, max_width, max_height):
    res_width = max_width
    mul = max_width / image_width
    res_height = mul * image_height

    if res_height > max_height:
        mul = max_height / image_height
        res_width = mul * image_width

    return {
        "width": res_width,
        "height": res_height,
        "factor": mul
    }



from datetime import datetime

from dateutil.parser import parse


def parse_visit_info(visit_information):
    parsed_info = {
        'doctor_name': visit_information.get('doctorName', ''),
        'location': visit_information.get('locationName', ''),
        'visit_reason': visit_information.get('detailVisitReason', ''),
        'formatted_date': '',
        'formatted_time': '',
    }

    kiosk_start = visit_information.get('kioskStartTime', '')
    if kiosk_start:
        kiosk_start_obj = parse(kiosk_start)
        parsed_info['formatted_date'] = datetime.strftime(kiosk_start_obj, '%Y-%m-%d')
        parsed_info['formatted_time'] = datetime.strftime(kiosk_start_obj, '%H:%M')

    return parsed_info


def parse_patient_info(patient_information):
    parsed_info = {
        'first_name': patient_information.get('firstName', '').capitalize(),
        'last_name': patient_information.get('lastName', '').capitalize(),
        'middle_name': patient_information.get('middleName', '').capitalize(),
        'gender': patient_information.get('gender', '').capitalize(),
        'birth_date': '',
        'ethnicity': '',
    }

    ethnicities = patient_information.get('ethnicity', '')
    if isinstance(ethnicities, list):
        ethnicities = ', '.join(ethnicities)
    parsed_info['ethnicity'] = str(ethnicities)

    birth_date = patient_information.get('birthDate')
    if birth_date:
        birth_date_obj = datetime.strptime(birth_date, '%Y-%m-%d')
        diff = datetime.today() - birth_date_obj
        parsed_info['age_years'] = str(int(diff.days / 365.25))
        parsed_info['birth_date'] = datetime.strftime(birth_date_obj, '%Y-%m-%d')

    name_list = [parsed_info['first_name'], parsed_info['last_name']]

    middle_name = parsed_info.get('middle_name', '')
    if middle_name:
        name_list.insert(1, '%s.' % middle_name[0].capitalize())

    parsed_info['full_name'] = ' '.join(name_list)

    return parsed_info, patient_information.get('contactInformation', {})



from io import BytesIO

from PIL import Image as PILImage
from reportlab.lib.enums import TA_CENTER
from reportlab.lib.pagesizes import letter
from reportlab.lib.styles import StyleSheet1, ParagraphStyle
from reportlab.platypus import SimpleDocTemplate, Image, Paragraph, Spacer

from .utils import stretch_image_size


class PDFSlidesGenerator:
    """
    This class is used for generation of PDF from measurements images
    """
    def __init__(self):
        self._buffer = BytesIO()
        self.doc = SimpleDocTemplate(self._buffer, pagesize=letter)
        self.frame_width = self.doc.width - 12  # 12 is default indent here
        self.frame_height = self.doc.height - 12
        self.elements = []
        self.styles = StyleSheet1()
        self.set_styles()

    def set_styles(self):
        self.styles.add(
            ParagraphStyle(name='normal', fontSize=18))
        self.styles.add(
            ParagraphStyle(name='title', fontSize=20, alignment=TA_CENTER))

    def add_spacer(self, height=12):
        self.elements.append(Spacer(self.frame_width, height))

    @staticmethod
    def _generate_image(stream, prefered_width, max_height):
        image = PILImage.open(stream)
        stream.seek(0)
        image_stretch = stretch_image_size(image.width, image.height, prefered_width, max_height)
        return Image(stream, image_stretch['width'], image_stretch['height'])

    def add_image(self, image):
        flowable_image = self._generate_image(image, self.frame_width, self.frame_height)
        self.elements.append(flowable_image)
        self.add_spacer()

    def add_caption(self, caption, style='normal'):
        caption_style = self.styles.get(style)
        flowable_caption = Paragraph(text=caption, style=caption_style)
        self.elements.append(flowable_caption)
        self.add_spacer()

    def render(self):
        self.doc.build(self.elements)
        self._buffer.seek(0)
        return self._buffer



import json
import unittest
from os.path import dirname, join

from dotenv import load_dotenv

from pdf_generator.generator import PDFGenerator

load_dotenv()


class PDFGeneratorTestCase(unittest.TestCase):

    @staticmethod
    def test_manual_json():
        file_name = 'test_jsons/input_pa_2343.json'
        kwargs = {
            "doctor_notes": json.loads(open(join(dirname(__file__), file_name)).read())
        }

        pdf_generator = PDFGenerator(**kwargs)
        pdf_buffer = pdf_generator.get_buffer()
        pdf_buffer_discharge = pdf_generator.get_buffer_discharge()

        output_file = file_name.replace('.json', '.pdf').replace('input', 'output').replace('test_jsons', '.')
        try:
            with open(output_file, 'wb') as file:
                file.write(pdf_buffer.getbuffer())
        except FileNotFoundError:
            pass

        try:
            with open(output_file.replace('output', 'discharge_output'), 'wb') as file:
                file.write(pdf_buffer_discharge.getbuffer())
        except FileNotFoundError:
            pass


if __name__ == "__main__":
    unittest.main()



from io import BytesIO

import requests

VALID_MEASUREMENTS = [
    'MEAN_ARTERIAL_PRESSURE',
    'DIASTOLIC',
    'SYSTOLIC',
    'PULSE',
    'BLOOD_OXYGEN',
    'WEIGHT',
    'HEIGHT',
    'TEMPERATURE',
    'RESPIRATORY_RATE'
]

IMAGE_FILE_EXTENSIONS = ['.jpg', '.jpeg', '.png']


def filter_measurements_with_picture(measurement):
    value = measurement.get('value', {})
    return (value.get('file') and
            isinstance(value.get('file'), str) and
            value.get('fileType', '') in IMAGE_FILE_EXTENSIONS and
            value.get('status', '') == 'AVAILABLE')


def get_measurements_pictures(measurements_list):
    detected_measurements = set()
    images = []
    for measurement in filter(filter_measurements_with_picture, measurements_list):
        value = measurement.get('value')
        friendly_body_part_name = '{} {}'.format(
            value.get('bodySide'),
            value.get('bodyPart')
        )
        if friendly_body_part_name not in detected_measurements:
            try:
                media_response = requests.get(value.get('file'))
                if media_response.ok:
                    images.append({
                        'title': friendly_body_part_name,
                        'media': BytesIO(media_response.content)
                    })
                    detected_measurements.add(friendly_body_part_name)
            except requests.ConnectionError:
                pass
    return images


def parse_measurements(measurements):
    parsed_measurements = {}
    parsed_media = []
    if measurements:
        for measurement in measurements:
            measurement_type = measurement.get('measureType', '')
            if measurement_type in VALID_MEASUREMENTS:
                parsed_measurements[measurement_type.lower()] = measurement.get('value')
        parsed_media = get_measurements_pictures(measurements)
    return parsed_measurements, parsed_media



import copy
import warnings

from io import BytesIO

from reportlab.lib import colors
from reportlab.lib.pagesizes import letter
from reportlab.lib.units import inch
from reportlab.pdfbase.pdfmetrics import stringWidth
from reportlab.platypus import SimpleDocTemplate, Paragraph, Spacer, Table, ListFlowable, ListItem, PageBreak, \
    TableStyle
from reportlab.platypus.flowables import HRFlowable, KeepTogether, Image
from PIL import Image as PILImage

from .preprocessor import JSONPreprocessor
from .styles import getStyleSheet, getTableStyles
from .utils import stretch_image_size


class PDFGenerator(JSONPreprocessor):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.buffer = BytesIO()
        self.buffer_discharge = BytesIO()
        self.stylesheet = getStyleSheet()
        self.tablestylesheet = getTableStyles()
        self._create_pdf()

    def get_buffer(self):
        return self.buffer

    def get_buffer_discharge(self):
        return self.buffer_discharge

    def _create_pdf(self):
        """
        The minus 6 below is because when the frame is applied to the doc,
        there is a padding that cannot be modified without creating an entire
        new class.  The hack is to offset it by 6, so now everything in the
        frame will be at the 1" margin.
        """
        bottom_margin = top_margin = left_margin = right_margin = 1.0 * inch - 6
        pdf_title = '{} {}'.format(self.patient_details.get('full_name'), self.visit_details.get('formatted_date'))
        pdf_args = {'pagesize': letter, 'rightMargin': right_margin, 'leftMargin': left_margin,
                    'topMargin': top_margin, 'bottomMargin': bottom_margin, 'showBoundary': 0,
                    'title': pdf_title}
        doc = SimpleDocTemplate(self.buffer, **pdf_args)
        doc_discharge = SimpleDocTemplate(self.buffer_discharge, **pdf_args)

        # this is for the doctors notes
        self.Elements = []
        # discharge
        self.Elements_discharge = []

        # precomputed self.Elements
        self.basic_horz_line = HRFlowable(color=colors.darkgrey, thickness=1)
        self.width_between_columns = Spacer(20, 1)

        ptext = 'Patient Visit Summary'
        self.Elements.append(Paragraph(ptext, self.stylesheet['title']))
        self.Elements.append(Spacer(1, 12))
        self.Elements_discharge.append(Paragraph(ptext, self.stylesheet['title']))
        self.Elements_discharge.append(Spacer(1, 12))

        visit_table_data = [
            [Paragraph('Doctor Name:', self.stylesheet['table_header']),
             self.width_between_columns,
             Paragraph(self.visit_details['doctor_name'], self.stylesheet['default'])],
            [Paragraph('Facility:', self.stylesheet['table_header']),
             self.width_between_columns,
             Paragraph(self.visit_details['location'], self.stylesheet['default'])],
            [Paragraph('Arrival Date:', self.stylesheet['table_header']),
             self.width_between_columns,
             Paragraph(self.visit_details['formatted_date'], self.stylesheet['default'])],
            [Paragraph('Arrival Time:', self.stylesheet['table_header']),
             self.width_between_columns,
             Paragraph(self.visit_details['formatted_time'], self.stylesheet['default'])],
            [self.Elements.append(Spacer(1, 12))],
            [Paragraph('Patient Name:', self.stylesheet['table_header']),
             self.width_between_columns,
             Paragraph(self.patient_details['full_name'], self.stylesheet['default'])],
            [Paragraph('Gender:', self.stylesheet['table_header']),
             self.width_between_columns,
             Paragraph(self.patient_details['gender'], self.stylesheet['default'])],
            [Paragraph('Ethnicity:', self.stylesheet['table_header']),
             self.width_between_columns,
             Paragraph(self.patient_details['ethnicity'], self.stylesheet['default'])],
            [Paragraph('Birth Date:', self.stylesheet['table_header']),
             self.width_between_columns,
             Paragraph(self.patient_details['birth_date'], self.stylesheet['default'])],
            [Paragraph('Age:', self.stylesheet['table_header']),
             self.width_between_columns,
             Paragraph(self.patient_details['age_years'], self.stylesheet['default'])],
            [self.Elements.append(Spacer(1, 12))],
            [Paragraph('Visit Reason:', self.stylesheet['table_header']),
             self.width_between_columns,
             Paragraph(self.visit_details['visit_reason'], self.stylesheet['default'])]
        ]

        visit_table = Table(
            data=visit_table_data, colWidths=self._get_max_width(visit_table_data),
            hAlign='LEFT', style=self.tablestylesheet['left'])

        self.Elements.append(visit_table)
        self.Elements.extend([Spacer(1, 12), self.basic_horz_line])

        self.Elements_discharge.append(visit_table)
        self.Elements_discharge.extend([Spacer(1, 12), self.basic_horz_line])

        patient_vital_group = []
        ptext = 'Patient Vitals'
        patient_vital_group.append(Paragraph(ptext, self.stylesheet['title']))
        patient_vital_group.append(Spacer(1, 12))
        if self.has_vitals_for_pdf():
            self._vitals_creation(patient_vital_group)
        else:
            patient_vital_group.append(Paragraph('No Vitals for this visit', self.stylesheet['default']))
        patient_vital_group.extend([Spacer(1, 12), self.basic_horz_line])
        self.Elements.append(KeepTogether(patient_vital_group))

        personal_medical_history = []
        ptext = 'Patient History'
        personal_medical_history.append(Paragraph(ptext, self.stylesheet['title']))
        personal_medical_history.append(Spacer(1, 12))

        ptext = '<u>Personal Medical History</u>'
        personal_medical_history.append(Paragraph(ptext, self.stylesheet['header']))

        personal_medical_history_data = []

        if len(self.patient_history.get('personal_history', [])) > 0 and \
                self.patient_history.get('personal_history') != ['']:
            for i in sorted(self.patient_history['personal_history']):
                personal_medical_history_data.append([Paragraph(i, self.stylesheet['table_header'])])
        else:
            personal_medical_history_data.append(
                [Paragraph('No Reported Personal Medical History', self.stylesheet['default'])]
            )

        personal_medical_history_table = Table(data=personal_medical_history_data,
                                               colWidths=self._get_max_width(personal_medical_history_data),
                                               hAlign='LEFT', style=self.tablestylesheet['top_align'])
        personal_medical_history.append(personal_medical_history_table)
        personal_medical_history.extend([Spacer(1, 12)])

        self.Elements.append(KeepTogether(personal_medical_history))

        surgical_history = []
        ptext = '<u>Past Surgical History</u>'
        surgical_history.append(Paragraph(ptext, self.stylesheet['header']))

        surgical_history_data = []

        surgical_history_items = sorted(self.patient_history.get('surgical_history', []))
        if len(surgical_history_items) > 0 and surgical_history_items != ['']:
            for i in surgical_history_items:
                surgical_history_data.append([Paragraph(i, self.stylesheet['table_header'])])
        else:
            surgical_history_data.append([Paragraph('No Reported Past Surgical History', self.stylesheet['default'])])

        surgical_history_table = Table(data=surgical_history_data, colWidths=self._get_max_width(surgical_history_data),
                                       hAlign='LEFT', style=self.tablestylesheet['top_align'])
        surgical_history.append(surgical_history_table)
        surgical_history.extend([Spacer(1, 12)])

        self.Elements.append(KeepTogether(surgical_history))

        family_history_group = []
        ptext = '<u>Family Medical History</u>'
        family_history_group.append(Paragraph(ptext, self.stylesheet['header']))

        family_history_data = []

        if self.patient_history.get('family_history', []):
            for i in sorted(self.patient_history['family_history']):
                family_history_data.append([Paragraph(i, self.stylesheet['table_header'])])
        else:
            family_history_data.append([Paragraph('No Reported Family Medical History', self.stylesheet['default'])])

        family_history_table = Table(data=family_history_data, colWidths=self._get_max_width(family_history_data),
                                     hAlign='LEFT', style=self.tablestylesheet['top_align'])

        family_history_group.append(family_history_table)
        family_history_group.extend([Spacer(1, 12), self.basic_horz_line])
        self.Elements.append(KeepTogether(family_history_group))

        patient_history_group = []
        patient_history_top_data = [[Paragraph('Medications', self.stylesheet['header_center']),
                                     Paragraph('Immunizations', self.stylesheet['header_center'])]]

        bullet_list = []
        for key in ['medications', 'immunizations']:

            end_list = []
            for item in self.patient_history[key]:
                end_list.append(ListItem(Paragraph(item, self.stylesheet['default_list'])))

            bullet_list.append(ListFlowable(end_list, bulletType='bullet', start=u'\u2022', leftIndent=8))

        patient_history_top_data.append(bullet_list)

        patient_history_top_table = Table(
            data=patient_history_top_data, colWidths=[doc.width / 2] * 2,
            style=self.tablestylesheet['top_align']
        )

        patient_history_group.append(patient_history_top_table)
        patient_history_group.append(Spacer(1, 12))

        patient_history_top_data = [[Paragraph('Medication Allergies', self.stylesheet['header_center']),
                                     Paragraph('Other Allergies', self.stylesheet['header_center'])]]

        bullet_list = []
        for key in ['medication_allergies', 'allergies']:

            end_list = []
            for item in sorted(self.patient_history[key]):
                end_list.append(ListItem(Paragraph(item, self.stylesheet['default_list'])))

            bullet_list.append(ListFlowable(end_list, bulletType='bullet', start=u'\u2022', leftIndent=8))

        patient_history_top_data.append(bullet_list)

        patient_history_top_table = Table(
            data=patient_history_top_data, colWidths=[doc.width / 2] * 2,
            style=self.tablestylesheet['top_align']
        )

        patient_history_group.append(patient_history_top_table)
        patient_history_group.append(Spacer(1, 12))
        self.Elements.append(KeepTogether(patient_history_group))

        social_history_group = []
        ptext = '<u>Social History</u>'
        social_history_group.append(Paragraph(ptext, self.stylesheet['header']))

        social_data = [[Paragraph('Smoking Status: ', self.stylesheet['table_header']),
                        self.width_between_columns,
                        Paragraph(self.patient_history['tobacco'], self.stylesheet['default'])],
                       [Paragraph('Alcohol Status: ', self.stylesheet['table_header']),
                        self.width_between_columns,
                        Paragraph(self.patient_history['alcohol'], self.stylesheet['default'])],
                       [Paragraph('Caffeine Status: ', self.stylesheet['table_header']),
                        self.width_between_columns,
                        Paragraph(self.patient_history['caffeine'], self.stylesheet['default'])]]

        social_table = Table(data=social_data, colWidths=self._get_max_width(social_data), hAlign='LEFT',
                             style=self.tablestylesheet['left'])

        social_history_group.append(social_table)
        social_history_group.append(Spacer(1, 12))
        self.Elements.append(KeepTogether(social_history_group))

        physical_exam_group = []
        ptext = 'Physical Exam'
        physical_exam_group.append(Paragraph(ptext, self.stylesheet['title']))
        physical_exam_group.append(Spacer(1, 12))

        if self.physical_exam:
            exam_data = [
                [Paragraph('{}: '.format(item['examName']), self.stylesheet['table_header']),
                 self.width_between_columns,
                 Paragraph(item['examResults'], self.stylesheet['default'])] for item in self.physical_exam
            ]
            exam_table = Table(data=exam_data, colWidths=self._get_max_width(exam_data), hAlign='LEFT',
                               style=self.tablestylesheet['exam_table'])
            physical_exam_group.append(exam_table)
        else:
            physical_exam_group.append(Paragraph('Physical Exam not conducted', self.stylesheet['default']))

        physical_exam_group.append(Spacer(1, 12))
        self.Elements.append(KeepTogether(physical_exam_group))

        self._symptoms_creation(self.Elements)

        self._add_ros_symptoms()

        # to align all the table columns, we will use a standard colwidths
        self.std_ill_col_widths = [(doc.width - 160) / 2, 80, (doc.width - 160) / 2, 80]
        self.std_trt_col_widths = [90, doc.width - 90]

        lab_group = self._lab_creation()
        self.Elements.append(KeepTogether(lab_group))

        # if no illnesses are selected (pdf generated during doctor using the app)
        # dont display any illness/treatment info.
        if len(self.illnesses_selected) > 0:
            self.symptoms_with_icds = list(map(list, set(map(tuple, self.symptoms_with_icds))))
            self._illness_creation(dta=self.illnesses_selected, ttl='Illnesses Diagnosed',
                                   element_target=self.Elements, symp_data=self.symptoms_with_icds, is_selected=True)
            if self.treatments_selected_formatted:
                self._treatment_creation(self.treatments_selected_formatted, 'Treatments', element_target=self.Elements,
                                         is_selected=True)

            # for the discharge, we dont want all the contributors and we can combine the illnesses and symptoms.
            # Must be run after the notes

            for i in self.illnesses_selected:
                ill_list = ['', i['icd_cd'], i['icd_desc']]
                self.symptoms_with_icds.insert(0, ill_list)

            self._illness_creation(ttl='Illnesses Diagnosed', element_target=self.Elements_discharge,
                                   symp_data=self.symptoms_with_icds)
            if self.treatments_selected_formatted:
                self._treatment_creation(
                    self.treatments_selected_formatted, 'Treatments',
                    element_target=self.Elements_discharge
                )

        add_doctors_notes = self.create_additional_notes('notes', 'Summary Doctor\'s Notes', [])
        self.Elements.append(KeepTogether(add_doctors_notes))

        if self.media:
            self.Elements.append(PageBreak())
            ptext = 'Supplemental Information'
            self.Elements.append(Paragraph(ptext, self.stylesheet['title']))
            self.Elements.append(Spacer(1, 12))
            self.Elements.append(self.basic_horz_line)
            self.Elements.append(Spacer(1, 12))
            if self.media:
                self._measurement_images(self.media, self.Elements)
        doc.build(self.Elements, onFirstPage=self._footer, onLaterPages=self._footer)
        doc_discharge.build(self.Elements_discharge, onFirstPage=self._footer, onLaterPages=self._footer)

    def _vitals_creation(self, target_elements):
        vital_table_data = [
            [Paragraph('Blood Pressure:', self.stylesheet['table_header']),
             Paragraph(self.measurements['blood_pressure'], self.stylesheet['default'])],
            [Paragraph('Pulse:', self.stylesheet['table_header']),
             Paragraph(self.measurements['pulse'], self.stylesheet['default'])],
            [Paragraph('Oxygen Saturation:', self.stylesheet['table_header']),
             Paragraph(self.measurements['blood_oxygen'], self.stylesheet['default'])],
            [Paragraph('Mean Arterial Pressure:', self.stylesheet['table_header']),
             Paragraph(self.measurements['mean_arterial_pressure'], self.stylesheet['default'])],
            [Paragraph('Weight:', self.stylesheet['table_header']),
             Paragraph(self.measurements['weight'], self.stylesheet['default'])],
            [Paragraph('Temperature:', self.stylesheet['table_header']),
             Paragraph(self.measurements['temperature'], self.stylesheet['default'])],
            [Paragraph('Height:', self.stylesheet['table_header']),
             Paragraph(self.measurements['height'], self.stylesheet['default'])],
            [Paragraph('Respiratory Rate:', self.stylesheet['table_header']),
             Paragraph(self.measurements['respiratory_rate'], self.stylesheet['default'])],
        ]

        vital_table = Table(data=self._table_splitter(vital_table_data), hAlign='LEFT',
                            style=self.tablestylesheet['left'])
        target_elements.append(vital_table)

    def _symptoms_creation(self, target_elements):
        symptom_group = []
        ptext = 'History of Present Illness'
        symptom_group.append(Spacer(1, 12))
        symptom_group.append(Paragraph(ptext, self.stylesheet['title']))
        symptom_group.append(Paragraph(self.doctor_notes.get('summaryHPI', ""), self.stylesheet['default']))
        symptom_group.append(Spacer(1, 12))

        presenting_symptoms = sorted(self.symptoms['presenting'])
        not_presenting_symptoms = sorted(self.symptoms['not_presenting'])

        if presenting_symptoms or not_presenting_symptoms:
            ptext = '<u>Symptoms Presenting</u>'
            symptom_group.append(Paragraph(ptext, self.stylesheet['header']))
            presenting_symptoms_data = []
            for i in presenting_symptoms:
                if len(i) == 1:
                    presenting_symptoms_data.append([Paragraph(i[0], self.stylesheet['table_header'])])
                else:
                    presenting_symptoms_data.append([Paragraph(i[0], self.stylesheet['sub_table_header']),
                                                     self.width_between_columns,
                                                     Paragraph(i[1], self.stylesheet['default'])])

            if len(presenting_symptoms_data) > 0:
                presenting_symptoms_table = Table(data=presenting_symptoms_data,
                                                  colWidths=self._get_max_width(presenting_symptoms_data),
                                                  hAlign='LEFT', style=self.tablestylesheet['symptoms'])
                symptom_group.append(presenting_symptoms_table)
            else:
                symptom_group.append(Paragraph('No Presenting Symptoms Reported', self.stylesheet['default']))

            symptom_group.append(Spacer(1, 12))
            target_elements.append(KeepTogether(symptom_group))

            non_presenting_symptom_group = []
            ptext = '<u>Symptoms Not Presenting</u>'
            non_presenting_symptom_group.append(Paragraph(ptext, self.stylesheet['header']))
            if not_presenting_symptoms:
                not_presenting_symptoms_data = ListFlowable([Paragraph(i, self.stylesheet['default_list'])
                                                             for i in not_presenting_symptoms],
                                                            bulletType='bullet', start=u'\u2022', leftIndent=10)
                non_presenting_symptom_group.append(not_presenting_symptoms_data)
            else:
                non_presenting_symptom_group.append(Paragraph('No Non Presenting Symptoms Reported',
                                                    self.stylesheet['default']))
            non_presenting_symptom_group.extend([Spacer(1, 12)])
            target_elements.append(KeepTogether(non_presenting_symptom_group))
        else:
            symptom_group.append(Paragraph('No symptoms collected for this visit', self.stylesheet['default']))
            target_elements.append(KeepTogether(symptom_group))

    def _lab_creation(self):
        lab_group = [Spacer(1, 12), Paragraph('Labs', self.stylesheet['title']), Spacer(1, 12)]

        lab_data = []
        for lab in self.symptoms['labs']:
            lab_data.append([Paragraph(lab[3], self.stylesheet['sub_table_header']),
                            self.width_between_columns,
                            Paragraph(lab[2], self.stylesheet['default'])])

        if not lab_data:
            lab_group.append(Paragraph('No labs for this visit', self.stylesheet['default']))
        else:
            lab_table = Table(data=lab_data, colWidths=self._get_max_width(lab_data), hAlign='LEFT',
                              style=self.tablestylesheet['symptoms'])
            lab_group.append(lab_table)
        lab_group.append(Spacer(1, 12))

        return lab_group

    def _medication_instructions_creation(self, ttl, element_target):
        self.create_additional_notes(
            'medication_instructions',
            ttl,
            element_target
        )
        element_target.append(Spacer(1, 12))

    def _illness_creation(self, ttl, element_target, dta=None, symp_data=None, is_selected=False):
        illnesses_diagnosed = [Spacer(1, 12), Paragraph(ttl, self.stylesheet['title'])]
        if is_selected:
            illnesses_diagnosed = self.create_additional_notes('diagnostics', 'Diagnostic Notes', illnesses_diagnosed)
            # Add primary illness
            if dta:
                primary_illnesses = [i for i in dta if i['is_primary']]
                other_illnesses = copy.deepcopy(dta)
                if primary_illnesses:
                    p = primary_illnesses[0]
                    combined = [[Paragraph('Primary Diagnosis  -  ' + p['icd_cd'] + "  -  " + p['icd_desc'],
                                           self.stylesheet['bold'])]]
                    illnesses_diagnosed.append(Table(data=combined, colWidths=[480], hAlign='LEFT'))
                    illnesses_diagnosed.append(Spacer(1, 12))
                    other_illnesses.remove(p)

                if other_illnesses:
                    illnesses_diagnosed.append(Table(data=[[Paragraph("Other Diagnosis", self.stylesheet['bold'])]],
                                                     colWidths=[480], hAlign='LEFT'))
                    for d in other_illnesses:
                        illnesses_diagnosed.append(Table(data=[[Paragraph(d['icd_cd'] + '  -  ' + d['icd_desc'],
                                                                          self.stylesheet['default'])]],
                                                         colWidths=[480], hAlign='LEFT'))

        illnesses_diagnosed.append(Spacer(1, 12))

        if symp_data:
            for i in symp_data:
                combined = [[Paragraph(i[1] + ':    ' + i[2], self.stylesheet['default'])]]
                illnesses_diagnosed.append(Table(data=combined, colWidths=[480], hAlign='LEFT'))
            if not dta:
                illnesses_diagnosed.append(Spacer(1, 12))

        illnesses_diagnosed.append(Spacer(1, 18))

        if dta:
            for n, i in enumerate(dta):
                current_ill = []
                combined = [[Paragraph("{}     {}".format(i['icd_cd'], i['icd_desc']),
                                       self.stylesheet['bold'])]]
                current_ill.append(Table(data=combined, colWidths=[480], hAlign='LEFT'))
                current_ill.append(Paragraph('Contributors', self.stylesheet['header_center']))
                current_ill.append(Spacer(1, 12))
                contrib_data = []
                if len(i['contributors']) == 0:
                        non_contrib = Paragraph('Contributors not available for doctor added illnesses.',
                                                self.stylesheet['default_center'])
                        current_ill.append(non_contrib)
                        illnesses_diagnosed.append(KeepTogether(current_ill))
                else:
                    for j in i['contributors']:
                        if ' - Not Presenting' in j[0]:
                            contrib_data.append([Paragraph(j[0], self.stylesheet['italic']),
                                                 Paragraph(j[1], self.stylesheet['default_number_list'])])
                        else:
                            contrib_data.append([Paragraph(j[0], self.stylesheet['default']),
                                                 Paragraph(j[1], self.stylesheet['default_number_list'])])
                    contrib_split = self._table_splitter(contrib_data)
                    current_ill.append(Table(
                        data=contrib_split, colWidths=self.std_ill_col_widths,
                        hAlign='LEFT', style=self.tablestylesheet['left']
                    ))
                    illnesses_diagnosed.append(KeepTogether(current_ill))

                illnesses_diagnosed.append(Spacer(1, 12))
                illnesses_diagnosed.append(Spacer(1, 12))

        illnesses_diagnosed.append(self.basic_horz_line)
        element_target.extend(illnesses_diagnosed)

    def _treatment_creation(self, dta, ttl, element_target, is_selected=False):
        for n, i in enumerate(dta):
            treatments_assigned = []
            if n == 0:
                ptext = ttl
                treatments_assigned.append(Paragraph(ptext, self.stylesheet['title']))
                if is_selected:
                    notes_title = 'Treatments Notes'
                    treatments_assigned = self.create_additional_notes('treatments', notes_title, treatments_assigned)
                treatments_assigned.append(Spacer(1, 12))
                self._medication_instructions_creation('Medication Instructions', treatments_assigned)
            ptext = '<u>' + i['type'] + '</u>'
            treatments_assigned.append(Paragraph(ptext, self.stylesheet['header']))

            for j in i['names']:
                if i['type'] in ['Prescription Drugs', 'OTC Drugs']:

                    name_data = [[
                        Paragraph(j['name'], self.stylesheet['header_center']),
                        Paragraph('', self.stylesheet['default'])
                    ]]
                    if 'quantity' in j and j['quantity'] is not None:
                        name_data.append([
                            Paragraph('Quantity', self.stylesheet['sub_table_header']),
                            Paragraph(str(j['quantity']), self.stylesheet['default'])
                        ])
                    if 'strength' in j and j['strength'] is not None:
                        name_data.append([
                            Paragraph('Strength', self.stylesheet['sub_table_header']),
                            Paragraph(j['strength'], self.stylesheet['default'])
                        ])
                    if 'directionsString' in j and j['directionsString'] is not None:
                        name_data.append([
                            Paragraph('Directions', self.stylesheet['sub_table_header']),
                            Paragraph(j['directionsString'], self.stylesheet['default'])
                        ])
                    name_data.append([
                        Paragraph('Used to Treat', self.stylesheet['sub_table_header']),
                        Paragraph(', '.join(j['illnesses']), self.stylesheet['default'])
                    ])
                else:
                    name_data = [
                        [
                            Paragraph(j['name'], self.stylesheet['header_center']),
                            Paragraph('', self.stylesheet['default'])
                        ], [
                            Paragraph('Used to Treat', self.stylesheet['sub_table_header']),
                            Paragraph(', '.join(j['illnesses']), self.stylesheet['default'])
                        ]
                    ]

                treatments_assigned.append(Table(data=name_data, colWidths=self.std_trt_col_widths,
                                                 hAlign='LEFT', style=self.tablestylesheet['treatment_table']))

                treatments_assigned.append(Spacer(1, 12))
            treatments_assigned.append(Spacer(1, 12))
            element_target.append(KeepTogether(treatments_assigned))

    @staticmethod
    def _footer(canvas, doc):
        foot_font = 'Roboto'
        foot_font_size = 10
        foot_text = "Page %d" % doc.page

        foot_line_height = doc.bottomMargin * 0.5 + foot_font_size
        canvas.saveState()
        canvas.setFont(foot_font, foot_font_size)
        canvas.drawString(doc.width + doc.leftMargin - canvas.stringWidth(foot_text, foot_font, foot_font_size),
                          doc.bottomMargin * 0.5, foot_text)
        canvas.line(doc.leftMargin, foot_line_height, doc.width + doc.leftMargin, foot_line_height)
        canvas.restoreState()

    @staticmethod
    def _table_splitter(chart):
        num_rows = len(chart)
        num_cols = len(chart[0])
        left_side = int(num_rows / 2) + (num_rows % 2 > 0)
        new_chart = []
        for i in range(left_side):
            new_row = []
            for j in range(num_cols):
                new_row.append(chart[i][j])
            if i + left_side < num_rows:
                for k in range(num_cols):
                    new_row.append(chart[i + left_side][k])
            new_chart.append(new_row)
        return new_chart

    @staticmethod
    def _get_max_width(table):
        # update if doc width changed
        docwidth = 480
        table_width = len(table[0])
        max_values = [0] * table_width
        curr_val = 0
        for i in range(len(table)):
            for j in range(len(table[i])):
                if isinstance(table[i][j], Paragraph):

                    if len(table[i][j].frags) > 0:
                        additional_indent = table[i][j].style.leftIndent + table[i][j].style.rightIndent
                        curr_val = (stringWidth(table[i][j].text, table[i][j].frags[0].fontName,
                                                table[i][j].frags[0].fontSize) +
                                    additional_indent)

                if isinstance(table[i][j], Spacer):
                    curr_val = table[i][j].width
                if curr_val > max_values[j]:
                    max_values[j] = curr_val

        if sum(max_values) > docwidth:
            max_vl = max(max_values)
            max_vl_i = max_values.index(max_vl)
            max_values_dupe = list(max_values)
            max_values_dupe.remove(max_vl)
            new_val = docwidth - sum(max_values_dupe)
            max_values[max_vl_i] = new_val
        return max_values

    def _measurement_images(self, media_objects, target_list):
        for media in media_objects:
            image = self._generate_image(media['media'], 4 * inch, 4 * inch)
            if image:
                image_container = [
                    Paragraph(media['title'], self.stylesheet['title']),
                    image
                ]
                target_list.append(KeepTogether(image_container))
                target_list.append(Spacer(1, 24))
            else:
                warnings.warn("Invalid image type for {}".format(media['title']))

    @staticmethod
    def _generate_image(bytes_stream, max_width, max_height):
        try:
            image = PILImage.open(bytes_stream)
        except IOError:
            return None
        else:
            bytes_stream.seek(0)
            image_stretch = stretch_image_size(image.width, image.height, max_width, max_height)
            return Image(bytes_stream, image_stretch['width'], image_stretch['height'])

    def create_additional_notes(self, notes_type, text, target):
        notes = self.additional_info.get(notes_type, "")
        if notes is not None:
            notes = notes.replace("\n", "<br/>\n")
            text = '<u>{}</u>'.format(text)
            target.append(Paragraph(text, self.stylesheet['header']))
            target.append(Paragraph(notes, self.stylesheet['default']))
        return target

    def _add_ros_symptoms(self):
        # Display Review of Symptoms
        ros_group = [
            Spacer(1, 12),
            Paragraph("Review of Symptoms", self.stylesheet['title']),
            Spacer(1, 6),
        ]
        if self.ros_symptoms:
            ros_table_data = []
            for group in sorted(self.ros_symptoms.keys()):
                content = sorted(['(A) {}'.format(s[3]) for s in self.ros_symptoms[group] if s[1]])
                content += sorted(['(D) {}'.format(s[3]) for s in self.ros_symptoms[group] if not s[1]])
                ros_table_data.append(
                    [Paragraph('{}: '.format(group), self.stylesheet['table_header']),
                     self.width_between_columns,
                     Paragraph(', '.join(content), self.stylesheet['default'])],
                )

            ros_table = Table(
                data=ros_table_data, colWidths=self._get_max_width(ros_table_data),
                hAlign='LEFT', vAlign='TOP', style=self.tablestylesheet['left'])
            ros_table.setStyle(TableStyle([
                ('VALIGN', (0, 0), (-1, -1), 'TOP')
            ]))
            ros_group.append(ros_table)
            ros_group.append(Spacer(1, 6))
        else:
            ros_group.append(Paragraph('No symptoms for this visit', self.stylesheet['default']))

        ros_group.append(Spacer(1, 12))
        self.Elements.append(KeepTogether(ros_group))



import os.path

from reportlab.lib.styles import StyleSheet1, ParagraphStyle
from reportlab.lib.enums import TA_CENTER, TA_LEFT, TA_RIGHT
from reportlab.lib import colors
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.platypus import TableStyle

project_path = os.path.dirname(os.path.abspath(__file__))

pdfmetrics.registerFont(TTFont('Roboto', os.path.join(project_path, 'font', 'Roboto-Thin.ttf')))
pdfmetrics.registerFont(TTFont('Roboto-Italic', os.path.join(project_path, 'font', 'Roboto-ThinItalic.ttf')))
pdfmetrics.registerFont(TTFont('Roboto-Bold', os.path.join(project_path, 'font', 'Roboto-Regular.ttf')))


def getStyleSheet():
    """Returns a stylesheet object"""
    stylesheet = StyleSheet1()

    stylesheet.add(ParagraphStyle(
        name='default',
        fontName='Roboto',
        fontSize=10,
        leading=10,
        leftIndent=0,
        rightIndent=0,
        firstLineIndent=0,
        alignment=TA_LEFT,
        spaceBefore=0,
        spaceAfter=0,
        textColor=colors.black,
        backColor=None,
        wordWrap=None,
        borderWidth=0,
        borderPadding=0,
        borderColor=None,
        borderRadius=None,
        allowWidows=1,
        allowOrphans=0,
        textTransform=None,  # 'uppercase' | 'lowercase' | None
        endDots=None,
        splitLongWords=1))

    stylesheet.add(ParagraphStyle(
        name='bold',
        parent=stylesheet['default'],
        fontName='Roboto-Bold'))

    stylesheet.add(ParagraphStyle(
        name='italic',
        parent=stylesheet['default'],
        fontName='Roboto-Italic'))

    stylesheet.add(ParagraphStyle(
        name='default_list',
        parent=stylesheet['default'],
        leading=15))

    stylesheet.add(ParagraphStyle(
        name='default_number_list',
        parent=stylesheet['default'],
        alignment=TA_RIGHT,
        rightIndent=20,
        leftIndent=10))

    stylesheet.add(ParagraphStyle(
        name='title',
        parent=stylesheet['default'],
        alignment=TA_CENTER,
        fontSize=14,
        leading=20,
        spaceBefore=0,
        spaceAfter=0))

    stylesheet.add(ParagraphStyle(
        name='default_center',
        parent=stylesheet['default'],
        alignment=TA_CENTER))

    stylesheet.add(ParagraphStyle(
        name='header',
        parent=stylesheet['default'],
        alignment=TA_LEFT,
        fontSize=11,
        leading=20,
        spaceBefore=0,
        spaceAfter=0))

    stylesheet.add(ParagraphStyle(
        name='sub_table_header',
        parent=stylesheet['default'],
        leftIndent=7))

    stylesheet.add(ParagraphStyle(
        name='table_header',
        parent=stylesheet['default'],
        fontName='Roboto-Bold'))

    stylesheet.add(ParagraphStyle(
        name='header_center',
        parent=stylesheet['default'],
        fontName='Roboto-Bold',
        alignment=TA_CENTER))

    stylesheet.add(ParagraphStyle(
        name='default_bullet',
        parent=stylesheet['default'],
        bulletColor=colors.black,
        value='square',
        bulletFontSize=1,
        bulletIndent=2))

    return stylesheet


def getTableStyles():
    tablestyles = {'left': TableStyle([
        ('LEFTPADDING', (0, 0), (-1, -1), 0),
        ('RIGHTPADDING', (0, 0), (-1, -1), 0),
        ('ALIGN', (0, 0), (-1, -1), 'LEFT')
    ]), 'left_box': TableStyle([
        ('LEFTPADDING', (0, 0), (-1, -1), 0),
        ('RIGHTPADDING', (0, 0), (-1, -1), 0),
        ('ALIGN', (0, 0), (-1, -1), 'LEFT'),
        ('INNERGRID', (0, 0), (-1, -1), 0.25, colors.black),
        ('BOX', (0, 0), (-1, -1), 0.25, colors.black)
    ]), 'symptoms': TableStyle([
        ('LEFTPADDING', (0, 0), (-1, -1), 0),
        ('RIGHTPADDING', (0, 0), (-1, -1), 0),
        ('ALIGN', (0, 0), (-1, -1), 'LEFT')
    ]), 'top_align': TableStyle([
        ('LEFTPADDING', (0, 0), (-1, -1), 0),
        ('RIGHTPADDING', (0, 0), (-1, -1), 0),
        ('VALIGN', (0, 0), (-1, -1), 'TOP')
    ]), 'treatment_table': TableStyle([
        ('LEFTPADDING', (0, 0), (-1, -1), 5),
        ('RIGHTPADDING', (0, 0), (-1, -1), 0),
        ('VALIGN', (0, 0), (-1, -1), 'TOP'),
        ('GRID', (0, 0), (-1, -1), 0.5, colors.grey),
        ('SPAN', (0, 0), (1, 0)),
        ('BOX', (0, 0), (-1, -1), 1, colors.black)
    ]), 'exam_table': TableStyle([
        ('LEFTPADDING', (0, 0), (-1, -1), 0),
        ('RIGHTPADDING', (0, 0), (-1, -1), 0),
        ('ALIGN', (0, 0), (-1, -1), 'LEFT'),
        ('VALIGN', (0, 0), (0, -1), 'TOP'),
        ('LINEBELOW', (0, 0), (-1, -2), 0.25, colors.black),
    ])}

    return tablestyles



from collections import defaultdict
from datetime import datetime

from pdf_generator.preprocessor.parsers.triage import connect_to_gdb

CATEGORY_MAP = {
    'Medications': 'medications',
    'Allergies': 'allergies',
    'Medication Allergies': 'medication_allergies',
    'Immunizations': 'immunizations',
    'Surgical History': 'surgical_history',
    'Packs Per Day': 'packsPerDay',
    'Drinks Per Day': 'drinksPerDay',
    'Caffeine Drinks Per Day': 'caffeinePerDay',
    'Caffeine User': 'caffeineUser',
    'Alcohol User': 'alcoholUser'
}


SOCIAL_VALUES = ['Packs Per Day', 'Drinks Per Day', 'Caffeine Drinks Per Day', 'Caffeine User', 'Alcohol User']


GDB_Q_GET_ILLNESSNAME_BY_ICDS = """MATCH(ill:MITA_Illness) where ill.icd10Code in {} with ill 
                                    return ill.name, ill.icd10Code"""


def get_illness_name_by_icd_codes(icd_codes):
    graph_db = connect_to_gdb()
    illnesses = graph_db.query(q=GDB_Q_GET_ILLNESSNAME_BY_ICDS.format(icd_codes))
    return [ill[0] for ill in illnesses]


def get_correct_noun_form(noun, ending, number):
    return noun if str(number) == '1' else noun + ending


def parse_social_history(health_history):
    alcohol_desc = 'Patient did not report a last alcoholic drink.'
    tobacco_desc = 'Patient is not and has never been a smoker.'
    caffeine_desc = 'Patient does not report consuming caffeine.'

    alc_per_day = health_history.get('drinksPerDay', '')
    alcohol_noun = get_correct_noun_form('drink', 's', alc_per_day)
    last_drink = health_history.get('lastDrinkDate', '')
    drinking_end = health_history.get('drinkingEndDate', '')
    drinking_start = health_history.get('drinkingStartDate', '')
    packs_per_day = health_history.get('packsPerDay', '')
    smoking_noun = get_correct_noun_form('pack', 's', packs_per_day)
    smoking_start = health_history.get('smokingStartDate', '')
    smoking_end = health_history.get('smokingEndDate', '')
    caff_per_day = health_history.get('caffeinePerDay', '')

    if last_drink and not drinking_end:
        alcohol_desc = (
            'Patient\'s last alcoholic drink was {}. '
            'Patient reports drinking {} {} per day.'.format(
                last_drink, alc_per_day, alcohol_noun))
    elif drinking_start and drinking_end:
        alcohol_desc = (
            'Patient started drinking on {} and stopped drinking on {}. '
            'Patient reports drinking {} {} per day.'.format(
                drinking_start, drinking_end, alc_per_day, alcohol_noun))

    if smoking_start and smoking_end:
        tobacco_desc = (
            'Patient started smoking on {} and stopped smoking on {}. '
            'Patient reports smoking {} {} per day.'.format(
                smoking_start, smoking_end, packs_per_day, smoking_noun))
    elif smoking_start and not smoking_end:
        tobacco_desc = (
            'Patient has been a smoker since {}. '
            'Patient reports smoking {} {} per day.'.format(
                smoking_start, packs_per_day, smoking_noun))

    if caff_per_day:
        noun = get_correct_noun_form('drink', 's', caff_per_day)
        caffeine_desc = (
            'Patient reports drinking {} caffeine {} per day.'.format(
                caff_per_day, noun))

    return {
        'alcohol': alcohol_desc,
        'tobacco': tobacco_desc,
        'caffeine': caffeine_desc,
    }


def parse_health_history(health_history):
    parsed_health_history = defaultdict(list)
    parsed_health_history.update({
        'medications': [],
        'allergies': [],
        'medication_allergies': [],
        'immunizations': [],
    })
    history_items = health_history.pop('historyItem', [])

    previous_history_codes = []

    for item in history_items:
        category = item['historyType']
        history_key = CATEGORY_MAP.get(category)
        raw_value = item.get('historyItem', '')

        if category in SOCIAL_VALUES:
            health_history[history_key] = raw_value

        else:
            parsed_value = None

            if item['itemType'] == 'family':
                history_key = 'family_history'
                parsed_value = '{} - {}'.format(category, item['familyRelationship'].title())

            elif not history_key:
                history_key = 'personal_history'
                if item['symptomID'] == 'SYMPT0002070':     # <= BAK-2653
                    # Use the ones in the current year only.
                    if datetime.today().year == datetime.strptime(item['dateDetected'], '%Y-%m-%d').year:
                        previous_history_codes.append(raw_value)
                    continue

                parsed_value = category
                if raw_value:
                    parsed_value = '{} - {}'.format(parsed_value, raw_value)

            parsed_health_history[history_key].append(parsed_value or raw_value)

    illness_data = get_illness_name_by_icd_codes(previous_history_codes)
    for illness in illness_data:
        parsed_health_history['personal_history'].append('Previous Visit History - {}'.format(illness))

    parsed_health_history.update(**parse_social_history(health_history))

    return parsed_health_history


def clean_patient_history(patient_history):
    for key, value in patient_history.items():
        if not value:
            patient_history[key] = ['No {} Reported'.format(' '.join([v.capitalize() for v in key.split('_')]))]
        else:
            # Symptoms can be in both the triage and history sections leading to a duplication.
            # Doing a type check to not affect alcohol, tobacco, and caffeine data.
            if isinstance(value, list):
                new_values = []
                for v in value:
                    if isinstance(v, str):
                        if not any(v.lower() in s.lower() for s in new_values if isinstance(s, str)):
                            new_values.append(v)
                    elif v not in new_values:
                        new_values.append(v)
                patient_history[key] = new_values

    return patient_history



from collections import defaultdict

from .parsers.patient_data import parse_visit_info, parse_patient_info
from .parsers.health_history import parse_health_history, clean_patient_history
from .parsers.measurements import parse_measurements
from .parsers.triage import parse_triage


class JSONPreprocessor:

    def __init__(self, **kwargs):
        self.doctor_notes = kwargs['doctor_notes']
        self.symptom_data = kwargs.get('symptom_data', {})
        self.physical_exam = []
        self.treatments_selected_formatted = None
        self.media = kwargs.get('media', [])
        self._parse_doctor_notes()

    def _parse_doctor_notes(self):
        self.visit_details = parse_visit_info(self.doctor_notes.get('visitInformation', {}))
        self.patient_details, self.contact_information = parse_patient_info(
            self.doctor_notes.get('patientInformation', {}))

        self.measurements, additional_media = parse_measurements(self.doctor_notes.get('measurements', {}))
        self.media.extend(additional_media)

        self.patient_history = parse_health_history(self.doctor_notes.get('patientHealthHistory', {}))
        # BAK-2127: Use the History object and then look for the symptom when not is found.
        if len(self.patient_history.get('allergies', [])) == 0:
            for s in self.doctor_notes.get('triage', []):
                if s['symptomId'] == 'SYMPT0000044':
                    self.patient_history['allergies'] = [v[0] for v in s['values'] if v[0]]
                    break

        self.symptoms, self.symptoms_with_icds, self.measurements_taken, self.ros_symptoms = parse_triage(
            triage=self.doctor_notes.get('triage', []),
            patient_history=self.patient_history,
            patient_gender=self.patient_details['gender'].lower(),
            symptom_data=self.symptom_data,
        )

        # Diagnostic Engine
        self.illnesses_selected = []
        icd_illnesses_not_displayed = []
        if self.doctor_notes.get('diagnosticEngine'):
            for item in self.doctor_notes['diagnosticEngine']:
                start_dict = {'icd_cd': str(item.get('icd10', '')), 'icd_desc': str(item.get('icdName', '')),
                              'is_primary': item.get('isPrimary', False)}
                contrib = []
                min_value = 0.010
                if item.get('isSelected'):
                    for contributor in item.get('contributors', []):
                        # this means values whose contribution is 0 is not displayed.  This is intended.
                        contrib_value = float(contributor.get('contribution', 0))
                        if contrib_value > min_value or contrib_value < - min_value:
                            contrib.append([str(contributor.get('symptomName')), contrib_value * 100.0])
                    start_dict['contributors'] = contrib
                    self.illnesses_selected.append(start_dict)

        self.illnesses_selected.sort(key=lambda v: v['icd_cd'])

        for illness in self.illnesses_selected:
            test_dict = defaultdict(int)
            for key, value in illness.get('contributors'):
                if key in self.symptoms['not_presenting']:
                    test_dict[key + ' - Not Presenting'] += value
                else:
                    test_dict[key] += value
            # the 90 is because there is a threshold of 90 that is set by the DE.  Want to keep that threshold
            # this is pretty arbitrary
            for key, value in test_dict.items():
                if value > 90:
                    test_dict[key] = 90
                if value < -90:
                    test_dict[key] = -90
            confidence = sorted(test_dict.items(), key=lambda v: abs(v[1]), reverse=True)
            confidence = [[k, "{0:.1f}".format(v) + '%'] for (k, v) in confidence]
            illness['contributors'] = confidence

        self.patient_history = clean_patient_history(self.patient_history)

        # Treatment Engine
        treatments_selected = []
        # these treatments should not be printed out as they are the defaults
        TREATMENT_BLACKLIST = ['INCOMPLETE', 'Not Needed', 'Normal Diet', 'No Restrictions', 'None Needed',
                               'Immediately', 'Return as needed']

        if self.doctor_notes.get('treatmentEngine'):
            for item in self.doctor_notes['treatmentEngine']:
                if item.get('icdCode', '') not in icd_illnesses_not_displayed:
                    for treatment in item.get('treatments', []):
                        for detail in treatment.get('details', []):
                            start_dict = {}
                            if detail.get('name', '') not in TREATMENT_BLACKLIST and detail.get('isSelected'):
                                start_dict['type'] = treatment.get('type', '')
                                start_dict['name'] = detail.get('name', '')
                                start_dict['illness'] = item.get('icdDesc', '')
                                treatments_selected.append(start_dict)

        treatments_selected_formatted_unordered = []
        for i in set([i.get('type', '') for i in treatments_selected]):
            start_dict = {'type': i, 'names': []}
            for j in set([j.get('name', '') for j in treatments_selected if j.get('type', '') == i]):
                next_dict = {'name': str(j), 'illnesses': []}
                if start_dict.get('type', '') in ['Prescription Drugs', 'OTC Drugs']:
                    if self.doctor_notes.get('drugInformation'):
                        for d in self.doctor_notes['drugInformation']:
                            if d['drugName'] == next_dict.get('name', ''):
                                next_dict['route'] = str(d.get('route', None))
                                next_dict['unit'] = str(d.get('unit', None))
                                next_dict['strength'] = str(d.get('strength', None))
                                next_dict['quantity'] = str(d.get('quantity', None))
                                next_dict['directionsString'] = str(d.get('directionsString', None))

                for k in set([k.get('illness', '') for k in treatments_selected if k.get('name', '') == j]):
                    next_dict['illnesses'].append(k)
                start_dict['names'].append(next_dict)
            treatments_selected_formatted_unordered.append(start_dict)

        treatment_order = {
            'Prescription Drugs': 1,
            'OTC Drugs': 2,
            'Labs': 3,
            'Imaging': 4,
            'Procedures': 5,
            'Diet': 6,
            'Activity': 7,
            'Physical Therapy': 8,
            'Counseling': 9,
            'Wound Care': 10,
            'Physical Exam': 11,
            'Return to Work/School Status': 12,
            'Discharge Disposition': 13
        }

        treatments_selected_formatted_unordered_types = [
            x for x in treatments_selected_formatted_unordered
            if isinstance(x.get('type'), str) and 'names' in x]

        self.treatments_selected_formatted = sorted(treatments_selected_formatted_unordered_types,
                                                    key=lambda x: treatment_order.get(x.get('type', ''), 1))

        additional_info_source = self.doctor_notes.get('additionalInformation', {})
        self.additional_info = {
            'notes': additional_info_source.get('additionalDoctorNotes'),
            'treatments': additional_info_source.get('treatmentDoctorNotes'),
            'diagnostics': additional_info_source.get('diagnosticDoctorNotes'),
            'medication_instructions': additional_info_source.get('medicationInstructions')
        }
        self.physical_exam = additional_info_source.get('physicalExam', [])

        if self.measurements.get('systolic'):
            systolic_value = self.measurements.get('systolic', {}).get('value', '0')
        else:
            systolic_value = None

        if self.measurements.get('diastolic'):
            diastolic_value = self.measurements.get('diastolic', {}).get('value', '0')
        else:
            diastolic_value = None

        if systolic_value is not None and diastolic_value is not None:
            self.measurements['blood_pressure'] = str(systolic_value) + '/' + str(diastolic_value) + ' mmHg'
        else:
            self.measurements['blood_pressure'] = 'Not Measured'

        if self.measurements.get('weight'):
            wgt = round(self.measurements.get('weight').get('value', '0'), 1)
            self.measurements['weight'] = str(wgt) + ' lbs'
        else:
            self.measurements['weight'] = 'Not Measured'

        if self.measurements.get('pulse'):
            self.measurements['pulse'] = str(self.measurements.get('pulse', {}).get('value', '0')) + ' bpm'
        else:
            self.measurements['pulse'] = 'Not Measured'

        if self.measurements.get('temperature'):
            temp = round(self.measurements.get('temperature', {}).get('value', '0'), 1)
            self.measurements['temperature'] = str(temp) + ' °F'
        else:
            self.measurements['temperature'] = 'Not Measured'

        if self.measurements.get('mean_arterial_pressure'):
            map_value = self.measurements.get('mean_arterial_pressure', {}).get('value', '0')
            self.measurements['mean_arterial_pressure'] = str(int(float(map_value)))
        else:
            self.measurements['mean_arterial_pressure'] = 'Not Measured'

        if self.measurements.get('blood_oxygen'):
            map_value = self.measurements.get('blood_oxygen', {}).get('value', '0')
            self.measurements['blood_oxygen'] = str(int(float(map_value))) + ' %'
        else:
            self.measurements['blood_oxygen'] = 'Not Measured'

        if self.measurements.get('height'):
            map_value = self.measurements.get('height', {}).get('value', '0')
            self.measurements['height'] = str(map_value) + ' inch'
        else:
            self.measurements['height'] = 'Not Measured'

        if self.measurements.get('respiratory_rate'):
            self.measurements['respiratory_rate'] = str(self.measurements['respiratory_rate'].get('value', 0))
        else:
            self.measurements['respiratory_rate'] = 'Not Measured'

    def get_notes_parsing_result(self):
        def _clean_symptoms(symptoms):
            return [[item[0], item[1]] for item in symptoms]

        def _clean_illness(illness):
            return {
                'icd_cd': illness.get('icd_cd'),
                'icd_desc': illness.get('icd_desc'),
                'contributors': [item[0] for item in illness.get('contributors', [])]
            }

        presenting_symptoms = _clean_symptoms(self.symptoms.get('presenting', []))
        not_presenting_symptoms = self.symptoms.get('not_presenting', [])
        selected_illnesses = [_clean_illness(illness) for illness in self.illnesses_selected] + [
            {
                'icd_cd': symptom[1],
                'icd_desc': symptom[2],
                'contributors': []
            } for symptom in self.symptoms_with_icds]
        allowed_measurements = {
            'blood_pressure',
            'weight',
            'pulse',
            'mean_arterial_pressure',
            'blood_oxygen',
            'temperature'
        }
        vitals = {key: value for key, value in self.measurements.items() if key in allowed_measurements}
        return {
            'visitDetails': self.visit_details,
            'patientDetails': self.patient_details,
            'contactInformation': self.contact_information,
            'vitals': vitals,
            'patientHistory': self.patient_history,
            'physicalExam': self.physical_exam,
            'presentingSymptoms': presenting_symptoms,
            'notPresentingSymptoms': not_presenting_symptoms,
            'illnessesSelected': selected_illnesses,
            'treatmentsSelected': self.treatments_selected_formatted,
            'potentialMeasurements': self.measurements_taken,
            'additionalInformation': self.additional_info
        }

    def has_diagnoses_and_treatments(self):
        return self.illnesses_selected and self.treatments_selected_formatted

    @staticmethod
    def _has_measurement(key, parsed_measurements):
        return key in parsed_measurements and parsed_measurements[key].get("value", 0) > 0

    def has_vitals_for_pdf(self):
        measurements = parse_measurements(self.doctor_notes.get('measurements', {}))[0]
        vitals_keys = ["pulse", "blood_oxygen", "mean_arterial_pressure",
                       "weight", "temperature", "height", "respiratory_rate"]
        has_blood_pressure = all(self._has_measurement(value, measurements) for value in ["systolic", "diastolic"])
        has_other_measurements = any(self._has_measurement(key, measurements) for key in vitals_keys)
        return has_blood_pressure or has_other_measurements



from os import getenv
from neo4jrestclient.client import GraphDatabase
from pandas import DataFrame

from pdf_generator.utils import time_convert

SYMPTOM_BLACKLIST = [
    None,
    'SYMPT0000001',
    'SYMPT0000002',
    'SYMPT0000004',
    'SYMPT0000005',
    'SYMPT0000006',
    'SYMPT0000007',
    'SYMPT0000008',
    'SYMPT0000009',
    'SYMPT0000010',
    'SYMPT0000097',
    'SYMPT0000234',
    'SYMPT0000998',
    'SYMPT0001218',
    'SYMPT0001301',
    'SYMPT0009999',
]

POTENTIAL_MEASUREMENTS = [
    'SYMPT0001328',
    'SYMPT0001981',
    'SYMPT0001982',
    'SYMPT0001983',
    'SYMPT0002001',
    'SYMPT0002007',
    'SYMPT0002008',
]

# need to manually change some values for BP and diabetes:
bp_name_map = {
    'type_1': 'Type 1',
    'type_2': 'Type 2',
    'hyper_stage_1': 'Hypertension Stage 1',
    'hyper_stage_2': 'Hypertension Stage 2',
    'hyper_crisis': 'Hypertensive Crises',
    'hyper_crises': 'Hypertensive Crises',

}


GDB_Q_GET_SYMPTOM_ICDS = """Match(s:SymptomTemplate) where length(s.ICDRCode) > 1 with s optional 
                            Match (i: Illnesss{icd10Code : s.ICDRCode})  return s.code, s.name, s.ICDRCode, i.name"""


def connect_to_gdb():
    GDB_URL = getenv('GDB_URL')
    GDB_USER = getenv('GDB_USERNAME')
    GDB_PASS = getenv('GDB_PASSWORD')

    return GraphDatabase(GDB_URL, username=GDB_USER, password=GDB_PASS)


def get_symptom_icds():
    graph_db = connect_to_gdb()
    symptomICDLU_query = graph_db.query(q=GDB_Q_GET_SYMPTOM_ICDS)
    symptomICDLU = DataFrame(
        [i for i in symptomICDLU_query], columns=['symptomCode', 'symptomName', 'icdCode', 'icdName']
    )
    symptomICDLU.set_index('symptomCode', inplace=True, drop=True)

    return symptomICDLU


def parse_symptoms(symptoms):
    symptom_icds = get_symptom_icds()

    symptoms_with_icds = []

    presenting_symptoms = []
    non_presenting_symptoms = []

    symptom_group_order_template = ['General', 'Behaviour', 'behavior' 'Neurological', 'Measurements', 'Physical']
    symptom_keys = symptoms.keys()

    # Place existing symptoms first in list following order above
    symptom_group_order = [i for i in symptom_group_order_template if i in symptom_keys]
    # Add remaining present groups
    symptom_group_order.extend([i for i in symptoms.keys() if i not in symptom_group_order])

    for group in symptom_group_order:
        for symptom in symptoms.get(group, []):
            symptom_id, response, value, name, _ = symptom
            if not response:
                non_presenting_symptoms.append(name)
            else:
                # Capture symptoms with ICD codes
                if symptom_id in list(symptom_icds.index):
                    # Initial '' to keep same format as icds
                    symptoms_with_icds.append(
                        ['', symptom_icds.at[symptom_id, 'icdCode'], symptom_icds.at[symptom_id, 'symptomName']]
                    )

                presenting_symptoms.append([name, value, 'sub-heading'])

    return_symptoms = {
        'presenting': presenting_symptoms,
        'not_presenting': list(dict.fromkeys(non_presenting_symptoms)),
    }

    return return_symptoms, symptoms_with_icds


def is_skip_symptom(symptom_id, response, gender):
    # Even though response should be str, it comes in as bool all too often
    if response == 'skip':
        return True

    if symptom_id in SYMPTOM_BLACKLIST:
        return True

    # SYMPT0000213 = pregnancy
    if symptom_id == 'SYMPT0000213' and gender == 'male':
        return True


def parse_triage(triage, patient_history, patient_gender, symptom_data):
    measurements_taken = []
    symptom_ids = {'General': []}

    for item in triage:
        symptom_id = item.get('symptomId', '')

        response = str(item.get('response', '')).lower()
        # Negative response should be boolean to facilitate logic
        if response in ['false', '']:
            response = False

        if is_skip_symptom(symptom_id, response, patient_gender):
            continue

        values = item.get('values', [])

        symptom_name = item.get('symptomName', '')
        if symptom_name in bp_name_map.keys():
            symptom_name = bp_name_map[symptom_name]

        symptom_group = item.get('symptomGroup', '')

        if symptom_id in POTENTIAL_MEASUREMENTS:
            measurements_taken.append(symptom_name)

        # Parse health history categories
        lookup_category = '_'.join(item.get('categoryName', '').split(' ')).lower()
        if lookup_category in patient_history.keys() and response == 'true':
            if not any(symptom_name.lower() in s.lower() for s in patient_history[lookup_category]):
                patient_history[lookup_category].append(symptom_name)

        # Initialize group list if it does not exist.
        if symptom_group not in symptom_ids.keys():
            symptom_ids[symptom_group] = []

        normalized_symptom = [symptom_id, response, '', symptom_name,
                              symptom_data.get(symptom_id, {}).get('logicalGroupNames', [])  # ROS Group
                              ]
        # number = ''
        # descriptor = ''

        # TODO: Figure out if this comes from triage at all.
        # SYMPT0000009 = time since last drink, parse and continue
        if symptom_id == 'SYMPT0000009' and response:
            time = time_convert(values[0])
            normalized_symptom[2] = '{}{}'.format(time[0], time[1])
            if normalized_symptom not in symptom_ids[symptom_group]:
                symptom_ids[symptom_group].append(normalized_symptom)
        else:
            # values are an array of arrays because of legacy reasons, they have
            # been unpacked at this point so every symptom will have only one
            # array acting as a descriptor.
            # Labs will have descriptor in position 0, may be displayed with a
            # response of 'false' and must be displayed descriptor - value
            # while the rest of the symptoms are the opposite and false responses
            # are not described, only the name is displayed.
            if values and values[0]:
                for i, val in enumerate(values[0]):
                    if not isinstance(val, str):
                        values[0][i] = '' if val is None else str(val)
                lab_desc, number, descriptor = values[0]
                if symptom_group == 'Labs':
                    if not response:
                        normalized_symptom[1] = 'false'
                    measurement = '' if item.get('measurement') is None else str(item.get('measurement'))
                    if measurement and lab_desc and measurement != lab_desc:
                        normalized_symptom[2] = '{} - {}'.format(lab_desc, measurement)
                    else:
                        normalized_symptom[2] = lab_desc if lab_desc else measurement
                elif response:
                    if number and descriptor:
                        normalized_symptom[2] = '{} - {}'.format(number, descriptor)
                    else:
                        normalized_symptom[2] = descriptor if descriptor else number

        if normalized_symptom not in symptom_ids[symptom_group]:
            symptom_ids[symptom_group].append(normalized_symptom)

    # Removes labs before regular symptom parsing.
    lab_symptoms = symptom_ids.pop('Labs', [])

    symptoms, symptoms_with_icds = parse_symptoms(symptom_ids)

    symptoms['labs'] = lab_symptoms

    return symptoms, symptoms_with_icds, measurements_taken, _compose_ros_symptoms(symptom_ids, symptom_data, triage)


def _compose_ros_symptoms(symptom_ids, symptom_data, triage):
    """
    :param symptom_ids:
    :param symptom_data:
    :type symptom_data: dict
    :return:
    """
    ros_symptoms = {}

    def __add_to_ros_group(_group, _symptom, remove_basic=False):
        if remove_basic:
            _symptom[3] = _symptom[3].replace("Basic ", "")
        if _group in ros_symptoms.keys():
            ros_symptoms[_group].append(_symptom)
        else:
            ros_symptoms[_group] = [_symptom]

    # Get all symptoms as a list
    symptom_ids_list = []
    for g in symptom_ids:
        symptom_ids_list += symptom_ids[g]

    for symptom in symptom_ids_list:
        for g in symptom[4]:
            if str(g).startswith('ROS-'):
                __add_to_ros_group(g[4:], symptom)

        # BAK-2669
        symptom_id = symptom[0]
        _symptom_detail = symptom_data.get(symptom_id, {})
        symptom_name = _symptom_detail.get('symptomName', '')
        if symptom_name.startswith('Basic '):
            if _symptom_detail.get("symptomGroup") == "Pain/Swelling":
                if "pain" in symptom_name.lower():
                    # Get Pain Location symptom
                    for _id, s in symptom_data.items():
                        if s.get('categoryName') == _symptom_detail.get('categoryName') and \
                                'location' in s.get('symptomName', '').lower():
                            if _id in [ss.get('symptomId') for ss in triage]:
                                loc_symptom = [ss for ss in triage if ss.get('symptomId') == _id][0]
                                if loc_symptom['response'] and loc_symptom['values'] and loc_symptom['values'][0][0]:
                                    symptom[3] = symptom[3] + ", " + loc_symptom['values'][0][0]
                                    __add_to_ros_group('MusculoSkeletal', symptom, remove_basic=True)
                elif 'swelling' in symptom_name.lower():
                    __add_to_ros_group('MusculoSkeletal', symptom, remove_basic=True)
                else:
                    pass

        if "itch" in symptom_name:
            __add_to_ros_group('MusculoSkeletal', symptom, remove_basic=True)

    return ros_symptoms



#! /usr/bin/env python
import os
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pdf-generator-advinow',
    version='0.4.1',
    description='Utility for producing PDFs from Doctor Notes.',
    url='https://gitlab.com/AdviNow/pdf_generator/',
    download_url='https://gitlab.com/AdviNow/pdf_generator/-/archive/master/pdf_generator-master.tar.gz',
    install_requires=[
        'neo4jrestclient==2.1.1',
        'pandas==1.3.5',
        'python-dotenv==0.9.1',
        'reportlab==3.6.9',
        'requests==2.21.0',
        'restricted-pkg==1.1.2',
        'python-dateutil==2.7.3',
    ],
    license='MIT',
    include_package_data=True,
    author='AdviNow Medical',
    packages=['pdf_generator'],
)
